unit Keyb;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, StdCtrls, ComCtrls, TntForms;

type
  TKeybLay = class(TTntForm)
    StringGrid1: TStringGrid;
    Cancel: TButton;
    SaveChanges: TButton;
    LabHex: TLabel;
    LabDec: TLabel;
    KeybPatch: TCheckBox;
    Help: TButton;
    EnterChar: TRadioButton;
    Label1: TLabel;
    CharToEnter: TEdit;
    KeyCodeEdit: TEdit;
    Label2: TLabel;
    KeyCode: TUpDown;
    SpecFunc: TRadioButton;
    SpecFuncType: TComboBox;
    DoNothing: TRadioButton;
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure SaveChangesClick(Sender: TObject);
    procedure StringGrid1SetEditText(Sender: TObject; ACol, ARow: Integer;
      const Value: String);
    procedure StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure StringGrid1GetEditText(Sender: TObject; ACol, ARow: Integer;
      var Value: String);
    procedure KeybPatchClick(Sender: TObject);
    procedure StringGrid1Exit(Sender: TObject);
    procedure HelpClick(Sender: TObject);
    procedure DoNothingClick(Sender: TObject);
    procedure EnterCharClick(Sender: TObject);
    procedure SpecFuncClick(Sender: TObject);
    procedure SpecFuncTypeClick(Sender: TObject);
    procedure CharToEnterChange(Sender: TObject);
    procedure KeyCodeEditChange(Sender: TObject);
  private
    { Private declarations }
    noneChar:char;
    origPatched,updating:boolean;
    procedure SetMeaningControls(c:char; orig,shifted:boolean);
  public
    { Public declarations }
  end;

var
  KeybLay: TKeybLay;

implementation
uses FoProgi;
{$R *.DFM}

const KeybOffsets:array[1..5,1..5] of integer=
{The offset of the keyboard layout table is the same in every DOS version ($1edd), but there are some
 other needed offsets that differ. These are in order:
 the value used for the Exit game function ($e0 by default),
 the value used for keys without function ($ff by default),
 the code that determines wheter a value is special
 the value assigned to the Ctrl-C combination ($e0 by default, Ctrl-C means Exit game)
 the value assigned to the Ctrl-Alt-Del combination ($e0 by default, it means Exit game, too)}
      (($2ad43,$2c457,$2ad50,$2c46d,$2c783), {American DOS}
       ($2ad7d,$2c491,$2ad8a,$2c4a7,$2c7bd), {English DOS}
       ($2d123,$2e837,$2d130,$2e84d,$2eb63), {French DOS}
       ($2b903,$2d017,$2b910,$2d02d,$2d343), {German DOS}
       ($2d301,$2ea15,$2d30e,$2ea2b,$2ed41));{Spanish DOS}

var KeybPatchCode:array[1..16] of byte=
{Code to be inserted to change special values to $81..$9c}
    ($3c,$81,$72,$12,$3c,$9c,$77,$0e,$04,$60,$90,$90,$90,$90,$90,$90);
    {cmp al,81;jb +12;cmp al,9c;ja +e;add al,60;6*nop}

    OrigKeybCode:array[1..16] of byte=
{Original code, to undo the keyboard patch}
    ($3c,$f1,$72,$04,$3c,$fc,$76,$08,$3c,$e1,$72,$0a,$3c,$ec,$77,$06);
    {cmp al,f1;jb +4;cmp al,fc;jbe +8;cmp al,e1;jb +a;cmp al,ec;ja +6}

procedure TKeybLay.FormCreate(Sender: TObject);
var i:integer;
    s:String;
begin
 with StringGrid1 do
  begin
{We fill static cells}
   Cells[0,0]:='Key code';
   Cells[1,0]:='Key name';
   Cells[2,0]:='Original';
   Cells[3,0]:='Translated';
   Cells[4,0]:='Original shifted';
   Cells[5,0]:='Trans. shifted';
{We get key names from the keyboard driver}
   for i:=0 to 127 do
    begin
     Cells[0,i+1]:=IntToStr(i);
     SetLength(s,255);
     SetLength(s,GetKeyNameText(i shl 16,PChar(s),255));
     Cells[1,i+1]:=s;
    end;
  end;
end;

procedure TKeybLay.FormShow(Sender: TObject);
var keys:array[0..127] of char;
    i:cardinal;
    f:TMemoryStream;
    code:array[1..16] of byte;
begin
 if not (OrigType in DOSVersions) then
{We notify the user that (s)he won't be able to see the original keyboard layout}
  begin
   Application.MessageBox('The original executable you specified is a Windows version.'#10#13'These versions don''t have a keyboard table because'#10#13'they use the Windows keyboard layout'#10#13'You won''t be able to see the original keyboard layout.',
              'Information',MB_ICONINFORMATION or MB_OK);
   for i:=1 to 128 do
    begin
     StringGrid1.Cells[2,i]:='?';
     StringGrid1.Cells[4,i]:='?';
    end;
   origPatched:=False;
  end
 else
  begin
{If the original is a DOS version, we can load it}
   f:=TMemoryStream.Create;
   f.LoadFromFile(OrigExePath);
   f.Position:=$1edd;
{128 shifted assignments}
   f.ReadBuffer(keys,128);
   for i:=0 to 127 do
    StringGrid1.Cells[4,i+1]:=keys[i];
{128 normal assignments}
   f.ReadBuffer(keys,128);
   for i:=0 to 127 do
    StringGrid1.Cells[2,i+1]:=keys[i];
   f.Position:=KeybOffsets[TransType,3];
   f.ReadBuffer(code,16);
   origPatched:=false;
   for i:=1 to 16 do
  {If there's difference between code and OrigKeybCode then the keyboard patch is enabled}
    if code[i]<>OrigKeybCode[i] then
     begin
      origPatched:=true;
      Break;
     end;
   f.free;
  end;
{We load the translated layout}
 f:=TMemoryStream.Create;
 f.LoadFromFile(TransExePath);
 f.Position:=$1edd;
{128 shifted assignments}
 f.ReadBuffer(keys,128);
 for i:=0 to 127 do
  StringGrid1.Cells[5,i+1]:=keys[i];
{128 normal assignments}
 f.ReadBuffer(keys,128);
 for i:=0 to 127 do
  StringGrid1.Cells[3,i+1]:=keys[i];
{We load the special-value code}
 f.Position:=KeybOffsets[TransType,3];
 f.ReadBuffer(code,16);
 KeybPatch.OnClick:=nil; {We must unbind the OnClick handler because it'd activate when changing
                          the Checked property and would change assignment that shouldn't be changed yet}
{We assume that it's the original}
 KeybPatch.Checked:=false;
 noneChar:=#$ff; {so $ff means nothing}
 for i:=1 to 16 do
{If there's difference between code and OrigKeybCode then the keyboard patch is enabled}
  if code[i]<>OrigKeybCode[i] then
   begin
    KeybPatch.Checked:=true;
    noneChar:=#$90; {in this case $90 means nothing}
    Break;
   end;
 KeybPatch.OnClick:=KeybPatchClick; {We restore the event handler}
 f.free;
end;

procedure TKeybLay.SaveChangesClick(Sender: TObject);
var keys:array[0..127] of char;
    i:cardinal;
    f:TMemoryStream;
    b:byte;
begin
 f:=TMemoryStream.Create;
 f.LoadFromFile(TransExePath);
{We put all shifted key assignments into an array to make writing them easier}
 for i:=0 to 127 do
  keys[i]:=StringGrid1.Cells[5,i+1][1];
 f.Position:=$1edd;
 f.WriteBuffer(Keys,128);
{We do the same with normal assignments}
 for i:=0 to 127 do
  keys[i]:=StringGrid1.Cells[3,i+1][1];
 f.WriteBuffer(Keys,128);
 if KeybPatch.Checked then
  begin
   {If keyboard patch is enabled}
   {Exit game is $80}
   f.Position:=KeybOffsets[TransType,1];
   b:=$80;
   f.WriteBuffer(b,1);
   {Ctrl-C and Ctrl-Alt-Del means $80, too}
   f.Position:=KeybOffsets[TransType,4];
   f.WriteBuffer(b,1);
   f.Position:=KeybOffsets[TransType,5];
   f.WriteBuffer(b,1);
   {Value $90 indicates keys without function}
   f.Position:=KeybOffsets[TransType,2];
   b:=$90;
   f.WriteBuffer(b,1);
   {We write the patch code itself}
   f.Position:=KeybOffsets[TransType,3];
   f.WriteBuffer(KeybPatchCode,16);
  end
 else
  begin
   {If keyboard patch isn't enabled}
   {Exit game is $e0}
   f.Position:=KeybOffsets[TransType,1];
   b:=$E0;
   f.WriteBuffer(b,1);
   {Ctrl-C and Ctrl-Alt-Del means $e0, too}
   f.Position:=KeybOffsets[TransType,4];
   f.WriteBuffer(b,1);
   f.Position:=KeybOffsets[TransType,5];
   f.WriteBuffer(b,1);
   {Value $ff indicates keys without function}
   f.Position:=KeybOffsets[TransType,2];
   b:=$FF;
   f.WriteBuffer(b,1);
   {We write the original code because we don't know which code is there now}
   f.Position:=KeybOffsets[TransType,3];
   f.WriteBuffer(OrigKeybCode,16);
  end;
 f.SaveToFile(TransExePath);
 f.free;
 ModalResult:=mrOK;
end;

procedure TKeybLay.StringGrid1SetEditText(Sender: TObject; ACol,
  ARow: Integer; const Value: String);
begin
{A key assignment is changed}
 if Value<>'' then
  {If the cell is not empty, we leave only its last character - there should be only one
   character in every editable cell. Then we can show its code on the bottom}
  begin
   StringGrid1.Cells[ACol,ARow]:=Value[Length(Value)];
   LabHex.Caption:='Hex: '+IntToHex(ord(Value[1]),2);
   LabDec.Caption:='Dec: '+IntToStr(ord(Value[1]));
   SetMeaningControls(Value[1],not Odd(ACol),ACol>=4);
  end
 else
{The cell is empty - we don't show anything for its code}
  begin
   LabHex.Caption:='Hex:';
   LabDec.Caption:='Dec:';
   SetMeaningControls(#0,not Odd(ACol),ACol>=4);
  end;
end;

procedure TKeybLay.StringGrid1SelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
{If the old cell is empty, we fill it with the character that does nothing}
 With StringGrid1 do
  if Cells[Col,Row]='' then
   begin
    if Odd(Col) then
     Cells[Col,Row]:=noneChar
    else if origPatched then
     Cells[Col,Row]:=#$90
    else
     Cells[Col,Row]:=#$FF;
   end;
end;

procedure TKeybLay.StringGrid1GetEditText(Sender: TObject; ACol,
  ARow: Integer; var Value: String);
begin
{The user has selected a new cell - we refresh the code display}
 if Value='' then
  begin
   LabHex.Caption:='Hex:';
   LabDec.Caption:='Dec:';
   SetMeaningControls(#0,not Odd(ACol),ACol>=4);
  end
 else
  begin
   LabHex.Caption:='Hex: '+IntToHex(ord(Value[1]),2);
   LabDec.Caption:='Dec: '+IntToStr(ord(Value[1]));
   SetMeaningControls(Value[1],not Odd(ACol),ACol>=4);
  end;

end;

procedure TKeybLay.KeybPatchClick(Sender: TObject);
Var i:integer;
begin
{The keyboard patch setting is changed}
{We set nonechar}
 if KeybPatch.Checked then noneChar:=#$90
 else noneChar:=#$FF;
 with StringGrid1 do
  if KeybPatch.Checked then
{If Keyboard patch became enabled}
   for i:=1 to 128 do
    begin
     case Cells[3,i][1] of
      #$FF:Cells[3,i]:=#$90; {We change all $FF to $90}
      #$E0..#$FC:Cells[3,i]:=chr(ord(Cells[3,i][1])-$60); {And special characters $e0..$fc to $80..$9c}
     end;
     case Cells[5,i][1] of
     {The same with shifted assignments}
      #$FF:Cells[5,i]:=#$90;
      #$E0..#$FC:Cells[5,i]:=chr(ord(Cells[5,i][1])-$60);
     end;
    end
  else
{If Keyboard patch became disabled}
   for i:=1 to 128 do
    begin
     case Cells[3,i][1] of
      #$90:Cells[3,i]:=#$FF; {We change all $90 to $FF}
      #$80..#$8f,#$91..#$9C:Cells[3,i]:=chr(ord(Cells[3,i][1])+$60);
      {And special characters $80..$9c (except $90) to $e0..$fc}
     end;
     case Cells[5,i][1] of
     {The same with shifted assignments}
      #$90:Cells[5,i]:=#$FF;
      #$80..#$8f,#$91..#$9C:Cells[5,i]:=chr(ord(Cells[5,i][1])+$60);
     end;
    end;
end;

procedure TKeybLay.StringGrid1Exit(Sender: TObject);
begin
{If the selected cell is empty, we fill it with the character that does nothing}
 With StringGrid1 do
  if Cells[Col,Row]='' then
   Cells[Col,Row]:=noneChar;
end;

procedure TKeybLay.HelpClick(Sender: TObject);
begin
 Application.HelpCommand(HELP_CONTEXT,4);
end;

procedure TKeybLay.DoNothingClick(Sender: TObject);
begin
 if updating then Exit;
 With StringGrid1 do
  Cells[Col,Row]:=noneChar;
end;

procedure TKeybLay.SetMeaningControls(c: char; orig,shifted: boolean);
var patched:boolean;
begin
 updating:=true;
 EnterChar.Enabled:=not orig;
 SpecFunc.Enabled:=not orig;
 DoNothing.Enabled:=not orig;
 CharToEnter.Enabled:=not orig;
 KeyCodeEdit.Enabled:=not orig;
 KeyCode.Enabled:=not orig;
 SpecFuncType.Enabled:=not orig;
 if shifted then
  Label1.Caption:='This key, when shifted,'
 else
  Label1.Caption:='This key';
 CharToEnter.Text:=c;
 KeyCode.Position:=ord(c);
 SpecFuncType.ItemIndex:=-1;
 if orig then patched:=origPatched
 else patched:=KeybPatch.Checked;
 if patched then
  case c of
   #$90,#0: DoNothing.Checked:=true;
   #128..#140: begin
                SpecFunc.Checked:=true;
                SpecFuncType.ItemIndex:=ord(c)-128;
               end;
   #145..#156: begin
                SpecFunc.Checked:=true;
                SpecFuncType.ItemIndex:=ord(c)-145+13;
               end;
   else EnterChar.Checked:=true;
  end
 else
  case c of
   #$FF,#0: DoNothing.Checked:=true;
   #224..#236: begin
                SpecFunc.Checked:=true;
                SpecFuncType.ItemIndex:=ord(c)-224;
               end;
   #241..#252: begin
                SpecFunc.Checked:=true;
                SpecFuncType.ItemIndex:=ord(c)-241+13;
               end;
   else EnterChar.Checked:=true;
  end;
 updating:=false;
end;

procedure TKeybLay.EnterCharClick(Sender: TObject);
begin
 if updating then Exit;
 With StringGrid1 do
  Cells[Col,Row]:=' ';
end;

procedure TKeybLay.SpecFuncClick(Sender: TObject);
begin
 if updating then Exit;
 if SpecFuncType.ItemIndex=-1 then SpecFuncType.ItemIndex:=0;
 SpecFuncTypeClick(Self);
end;

procedure TKeybLay.SpecFuncTypeClick(Sender: TObject);
begin
 if updating then Exit;
 if SpecFuncType.ItemIndex=-1 then Exit;
 with StringGrid1,SpecFuncType do
  if KeybPatch.Checked then
   begin
    if ItemIndex<13 then
     Cells[Col,Row]:=chr(128+ItemIndex)
    else
     Cells[Col,Row]:=chr(145-13+ItemIndex);
   end
  else
   begin
    if ItemIndex<13 then
     Cells[Col,Row]:=chr(224+ItemIndex)
    else
     Cells[Col,Row]:=chr(241-13+ItemIndex);
   end
end;

procedure TKeybLay.CharToEnterChange(Sender: TObject);
begin
 if updating then Exit;
 with StringGrid1 do
  Cells[Col,Row]:=CharToEnter.Text;
end;

procedure TKeybLay.KeyCodeEditChange(Sender: TObject);
var a,v:integer;
begin
 if updating then Exit;
 Val(KeyCodeEdit.Text,v,a);
  if a=0 then
   with StringGrid1 do
    Cells[Col,Row]:=chr(v);
end;

end.
