unit runmkpttxt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, TntForms;

type
  TRunCompiler = class(TTnTForm)
    GiveUp: TButton;
    Label1: TLabel;
    procedure FormShow(Sender: TObject);
  private
    { Private declarations }
    procedure RunMkPtTxt(var Msg:TMessage); message WM_USER;
  public
    { Public declarations }
    CouldRunExe:boolean;
    ErrorCode:cardinal;
    ProgOutput:string;
    parameters:string;
  end;

var
  RunCompiler: TRunCompiler;

implementation

uses FoProgi;

{$R *.DFM}

type
  TPipeReader = class(TThread)
  private
    FPipeContents: string;
    FPipeHandle: THandle;
    FSuccessful: boolean;
  public
    property PipeContents: string read FPipeContents;
    property Successful: boolean read FSuccessful;
    constructor Create(PipeHandle: THandle);
    procedure Stop(timeOut:integer);
  protected
    procedure Execute; override;
  end;

procedure TRunCompiler.FormShow(Sender: TObject);
begin
 PostMessage(Handle,WM_USER,0,0);
end;

procedure TRunCompiler.RunMkPtTxt(var Msg: TMessage);
var startup:STARTUPINFO;
    process:PROCESS_INFORMATION;
    read_pipe,read_pipe_tmp,write_pipe,error_pipe:cardinal;
    security:SECURITY_ATTRIBUTES;
    pipeReader:TPipeReader;
begin
 Application.ProcessMessages;
 CouldRunExe:=false;
 ErrorCode:=0;
 {We create a pipe. The output or mkpttxt will be redirected to this pipe, so
  we can display it if necessary}
 with Security do
  begin
   nLength:=sizeof(security);
   lpSecurityDescriptor:=nil;
   bInheritHandle:=true;
  end;
 CreatePipe(read_pipe_tmp,write_pipe,@security,0);
 {create a duplicate for the error channel - the target may close stdout and
  stderr independently}
 DuplicateHandle(GetCurrentProcess,write_pipe,
                 GetCurrentProcess,@error_pipe,
                 0,true,DUPLICATE_SAME_ACCESS);
 {make the read end uninheritable - if mkpttxt inherited it, it would never be closed}
 DuplicateHandle(GetCurrentProcess,read_pipe_tmp,
                 GetCurrentProcess,@read_pipe,
                 0,false,DUPLICATE_CLOSE_SOURCE or DUPLICATE_SAME_ACCESS);
 {Now we run mkpttxt as a child process, with output redirected to our pipe
  and without showing its window}
 FillChar(startup,sizeof(startup),0);
 with startup do
  begin
   cb:=sizeof(startup);
   dwFlags:=STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
   wShowWindow:=SW_HIDE;
   hStdInput:=GetStdHandle(STD_INPUT_HANDLE);
   hStdOutput:=write_pipe;
   hStdError:=error_pipe;
  end;
 if not CreateProcess(PChar(ExtractFilePath(PatchPath)+'mkpttxt.exe'),
                      PChar(ExtractFilePath(PatchPath)+'mkpttxt.exe '+parameters),nil,nil,true,0,nil,
                      PChar(ExtractFilePath(PatchPath)),startup,process) then
  begin
   ErrorCode:=GetLastError;
   FileClose(write_pipe);
   FileClose(error_pipe);
   FileClose(read_pipe);
   ModalResult:=mrCancel;
   Exit;
  end;
 FileClose(write_pipe); {This handle is given to mkpttxt; we should release ours}
 FileClose(error_pipe);
 {we need a separate thread for reading the pipe, since you can't wait on pipes...}
 pipeReader:=TPipeReader.Create(read_pipe);
 {wait until mkpttxt finishes, but don't forget handling window messages}
 repeat
  case MsgWaitForMultipleObjects(1,process.hProcess,false,INFINITE,QS_ALLINPUT) of
   WAIT_OBJECT_0: CouldRunExe:=true; {The program has terminated}
   WAIT_OBJECT_0+1: Application.ProcessMessages; {There is a message in the quene}
   else begin
         ErrorCode:=GetLastError;
         CloseHandle(process.hProcess);
         CloseHandle(process.hThread);
         FileClose(read_pipe);
         pipeReader.Stop(500);
         pipeReader.Free;
         ModalResult:=mrCancel;
         Exit;
        end;
  end;
 until CouldRunExe or (ModalResult<>mrNone);
 progOutput:='';
 if WaitForSingleObject(pipeReader.Handle,250)=WAIT_TIMEOUT then
 begin
  pipeReader.Stop(250);
 end;
 if pipeReader.Successful then
  begin
   ProgOutput:=pipeReader.PipeContents;
   ModalResult:=mrOK;
  end
 else
  ModalResult:=mrCancel;
 pipeReader.Free;
 FileClose(read_pipe);
 GetExitCodeProcess(process.hProcess,ErrorCode);
 CloseHandle(process.hProcess);
 CloseHandle(process.hThread);
end;

{ TPipeReader }

constructor TPipeReader.Create(PipeHandle: THandle);
begin
  FPipeHandle:=PipeHandle;
  FSuccessful:=false;
  inherited Create(false);
end;

procedure TPipeReader.Execute;
var s:string[255];
    bytesRead: cardinal;
begin
  inherited;
  while not Self.Terminated and ReadFile(FPipeHandle,s[1],255,bytesRead,nil) do
   begin
    s[0]:=char(bytesRead);
    FPipeContents:=FPipeContents+s;
   end;
  if (GetLastError = ERROR_BROKEN_PIPE) then
   FSuccessful:=true;
end;

procedure TPipeReader.Stop(timeOut:integer);
begin
 Terminate;
 if WaitForSingleObject(Self.Handle,timeOut)=WAIT_TIMEOUT then
  TerminateThread(Self.Handle,0);
end;

end.
