unit ttdptxt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Grids, ExtCtrls, AppEvnts, TntGrids, TntForms, Contnrs;

type
  TInGameText = class(TTntForm)
    StringGrid1: TTntStringGrid;
    Help: TButton;
    Cancel: TButton;
    SaveAndCompile: TButton;
    PreviewTimer: TTimer;
    ApplicationEvents1: TApplicationEvents;
    SaveChanges: TButton;
    GRFLabel: TLabel;
    procedure FormResize(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PreviewTimerTimer(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ApplicationEvents1Activate(Sender: TObject);
    procedure ApplicationEvents1Deactivate(Sender: TObject);
    procedure SaveAndCompileClick(Sender: TObject);
    procedure StringGrid1SelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure StringGrid1KeyPress(Sender: TObject; var Key: Char);
    procedure StringGrid1Exit(Sender: TObject);
    procedure HelpClick(Sender: TObject);
    procedure SaveChangesClick(Sender: TObject);
  private
    { Private declarations }
    DatUpdated:boolean;
    PatchDir,TxtName,DatName,NewName,ExeName:string;
    function RunCompPrg(parameters:string):boolean;
    procedure RunInit(var msg:TMessage); message WM_USER;
    procedure LoadFromList(var msg:TMessage); message WM_USER+1;
    function ReadTexts(textList: TObjectList):boolean;
    procedure WriteTexts;
//    function CheckForNewFile:boolean;
    procedure WMSysCommand(var Msg:TWMSysCommand); message WM_SYSCOMMAND;
    class function UnQuoteString(quoted: string; out resultOrError: string): boolean;
    class procedure ParsePatchTextList(const output: string; textList: TObjectList);
    procedure LoadPatchTexts(textList: TObjectList);
  public
    OrigString:WideString;
    DestList: TObjectList;
    function CheckCError:boolean;
    function GetPatchTexts:TObjectList;
    { Public declarations }
  end;

var
  InGameText: TInGameText;

type
  TPatchText = class
   private
    FID: integer;
    FName: string;
    FOriginal: WideString;
    FTranslated: WideString;
    function isChanged: boolean;
   public
    constructor Create(ID: integer; name:string; origText: WideString);
    property ID: integer read FID;
    property Name: string read FName;
    property Original: WideString read FOriginal write FOriginal;
    property Translated: WideString read FTranslated write FTranslated;
    property Changed: boolean read isChanged;
  end;

implementation

uses runmkpttxt, FoProgi, CompError, previewf, StrUtils;

{$R *.DFM}

procedure TInGameText.FormResize(Sender: TObject);
var remainingSpace: integer;
begin
 with StringGrid1 do
  begin
   StringGrid1.ColWidths[0]:=50;
   remainingSpace:=ClientWidth-ColWidths[0]-ColWidths[1]-5;
   ColWidths[2]:=remainingSpace div 2;
   ColWidths[3]:=remainingSpace div 2;
  end;
end;

procedure TInGameText.FormShow(Sender: TObject);
begin
 if DestList=nil then
  PostMessage(Handle,WM_USER,0,0)
 else
  PostMessage(Handle,WM_USER+1,0,0); 
end;

function TInGameText.ReadTexts(textList: TObjectList):boolean;
var f:TextFile;
    s,s2,s3,id:string;
    ws:WideString;
//    i,j,k,p,value,error:integer;
    j,p:integer;

 function ShowError(const ErrStr:string):boolean;
  begin
   Result:=true;
   Case Application.MessageBox(PChar('There is an error in ttdpttxt.txt in line '+IntToStr(j)+':'#10#13+
                                     ErrStr+#10#13+'Do you want to ignore the error and continue?'#10#13+
                                     '(Choosing No will close the in-game translator window)'),
                                     'Error in data file',MB_ICONERROR or MB_YESNO) of
    ID_YES: Result:=false;
    ID_NO: Result:=true;
   end;
  end;

  function AddTranslatedByName(const textName: string; const newTranslated: WideString):boolean;
  var i:integer;
      new:TPatchText;
  begin
   Result:=false;
   for i:=0 to textList.Count-1 do
    with textList[i] as TPatchText do
     if name=textName then
      begin
       Translated:=newTranslated;
       Exit;
      end;
   new:=TPatchText.Create(-1,textName,'???');
   new.Translated:=newTranslated;
   textList.Add(new);
  end;

begin
 Result:=false;
 //StringGrid1.RowCount:=2;
 AssignFile(f,TxtName);
 Reset(f);
 {i:=0;} j:=0;
 while not Eof(f) do
  begin
   Readln(f,s);
   inc(j);
   if Trim(s)='' then continue;
   case s[1] of
    ' ',#9: s:=trim(s);
    ';': continue;
    '/': if (Length(s)>1) and (s[2]='/') then continue;
    else
     begin
      p:=Pos('=',s);
      if p=0 then
       begin
        if ShowError('Not an assigment (equals sign missing)') then Exit else continue;
       end;
      if s2<>'' then
       begin
        if (s2[length(s2)]=#0) then SetLength(s2,length(s2)-1);
        //StringGrid1.RowCount:=i+1;
        //StringGrid1.Cells[0,i]:=id;
        ws:=FoForm.TTDTextToWideString(s2);
        //ws:=FoForm.MakeCStyleString(ws);
        //StringGrid1.Cells[1,i]:=ws;
        if AddTranslatedByName(id,ws) then Exit;
        {inc(i);} s2:='';
       end;
      id:=Copy(s,1,p-1);
      s:=Copy(s,p+1,Length(s)-p);
      s:=Trim(s);
     end;
   end;
   if not UnQuoteString(s,s3) then
    begin
     if ShowError(s3) then Exit else s2:='-- Error --';
    end
   else
    begin
     s2:=s2+s3;
    end;
   {if (s='') or (s[1]<>'"') then
    begin
     if ShowError('Text not in quotes') then Exit else continue;
    end;
   Delete(s,1,1);
   while (Length(s)>0) and (s[length(s)]<>'"') do Delete (s,Length(s),1);
   if s='' then
    begin
     if ShowError('Text not in quotes') then Exit else continue;
    end;
   Delete(s,Length(s),1);
   k:=1;
   while k<=length(s) do
    begin
     if s[k]='\' then
      begin
       if k=length(s) then
        begin
         if ShowError('Backslash without control char at the end of the line') then Exit else goto EndOfLoop;
        end
       else
        case s[k+1] of
         '"': begin s2:=s2+'"'; inc(k,2); end;
         'n': begin s2:=s2+#13; inc(k,2); end;
         '\': begin s2:=s2+'\'; inc(k,2); end;
         else
          begin
           Val('$'+Copy(s,k+1,2),value,Error);
           if Error<>0 then
            begin
             if ShowError('Incorrect backslash-sequence') then Exit else goto EndOfLoop;
            end;
           if value=0 then
            begin
             StringGrid1.RowCount:=i+1;
             StringGrid1.Cells[0,i]:=id; id:='';
             ws:=FoForm.TTDTextToWideString(s2);
             ws:=FoForm.MakeCStyleString(ws);
             StringGrid1.Cells[1,i]:=ws;
             inc(i); s2:='';
            end
           else
            s2:=s2+chr(value);
           inc(k,3);
          end;
        end;
      end
     else
      begin
       s2:=s2+s[k]; inc(k);
      end;
    end;}
  end;
 if s2<>'' then
  begin
   if (s2[length(s2)]=#0) then SetLength(s2,length(s2)-1);
   //StringGrid1.RowCount:=i+1;
   //StringGrid1.Cells[0,i]:=id;
   ws:=FoForm.TTDTextToWideString(s2);
   //ws:=FoForm.MakeCStyleString(ws);
   //StringGrid1.Cells[1,i]:=ws;
   if AddTranslatedByName(id,ws) then Exit;
   s2:='';
  end;
 CloseFile(f);
 Result:=true;
end;

function TInGameText.RunCompPrg(parameters:string): boolean;
var LastDatDate:TDateTime;
    s,s2:string;
begin
 Result:=false;
 DatUpdated:=false;
 RunCompiler.parameters:=parameters;
 try
  if FileExists(DatName) then
   LastDatDate:=FileDateToDateTime(FileAge(DatName))
  else
   LastDatDate:=0;
  case RunCompiler.ShowModal of
   mrOK: begin
          Result:=true;
          DatUpdated:=FileExists(DatName) and
                      (FileDateToDateTime(FileAge(DatName))>LastDatDate);
         end;
   mrCancel: begin
              s2:='mkpttxt.exe';
              SetLength(s,300);
              SetLength(s,FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ARGUMENT_ARRAY,nil,
                        RunCompiler.ErrorCode,0,PChar(s),300,@s2));
              Application.MessageBox(PChar('An error occured while trying to run mkpttxt.exe:'#10#13+s),
                                     'Error',MB_ICONERROR or MB_ICONEXCLAMATION);
             end;
   end;
 except
  on E:Exception do
   begin
    Application.ShowException(E);
    Result:=false;
   end;
 end;

end;

procedure TInGameText.RunInit(var msg: TMessage);
var textList:TObjectList;
begin
 SaveAndCompile.Show;
 SaveChanges.Hide;
 GRFLabel.Hide;
 textList:=GetPatchTexts;
 if textList=nil then
  begin
   ModalResult:=mrCancel;
   Exit;
  end;
 LoadPatchTexts(textList);
 textList.Free;
end;

procedure TInGameText.FormCreate(Sender: TObject);
begin
 StringGrid1.Cells[0,0]:='Id';
 StringGrid1.Cells[1,0]:='Name';
 StringGrid1.Cells[2,0]:='Original';
 StringGrid1.Cells[3,0]:='Translated';
end;

procedure TInGameText.PreviewTimerTimer(Sender: TObject);
begin
 if not PreviewForm.Visible then PreviewTimer.Enabled:=false;
 With StringGrid1 do
  PreviewForm.PreviewText:=StringGrid1.Cells[Col,Row];
end;

procedure TInGameText.FormClose(Sender: TObject; var Action: TCloseAction);
begin
 PreviewTimer.Enabled:=false;
 SetWindowPos(PreviewForm.Handle,HWND_NOTOPMOST,0,0,0,0,SWP_NOSIZE or SWP_NOMOVE or SWP_NOACTIVATE);
end;

procedure TInGameText.WriteTexts;
var f:TextFile;
    s,s2,s3:WideString;
    ansis:string;
    i,j:integer;
begin
 AssignFile(f,TxtName);
 Rewrite(f);
 Writeln(f,'// TTDPatch in-game text strings');
 Writeln(f,'// This file was written by TTD Translator 2.2');
 Writeln(f,'// If mkpttxt.exe can''t compile this file, look for an updated version of');
 Writeln(f,'// TTD Translator at http://users.tt-forums.net/csaboka/TTDTrans/ ,');
 Writeln(f,'// or send an e-mail to the author of TTD Translator: csaboka@gmail.com');
 Writeln(f,'//');
 Writeln(f,'// TTDPatch and mkpttxt are Copyright (C) Josef Drexler');
 Writeln(f,'//');
 for i:=1 to (StringGrid1.RowCount)-1 do
  begin
   s:='';
   if StringGrid1.Cells[2,i]=StringGrid1.Cells[3,i] then s:='; ';
   s:=StringGrid1.Cells[1,i]+'="';
   s2:=StringGrid1.Cells[3,i];
   if FoForm.CompileCStyleString(s2,s3) then
    s2:=s3
   else
    s2:='';
   ansis:=FoForm.WideStringToTTDText(s2);
   for j:=1 to Length(ansis) do
    begin
     case ansis[j] of
      #0..#12,#14..#31,#$7B..#$9E: s:=s+'\'+IntToHex(ord(ansis[j]),2);
      #13: s:=s+'\n';
      '"': s:=s+'\"';
      '\': s:=s+'\\';
      else s:=s+ansis[j];
     end;
     if ((ansis[j]=' ') and (length(s)>70)) or (Length(s)>=80) then
      begin
       s:=s+'"';
       writeln(f,s);
       s:=#9'"';
      end;
    end;
   s:=s+'\00"';
   Writeln(f,s);
  end;
 CloseFile(f);
end;

procedure TInGameText.ApplicationEvents1Activate(Sender: TObject);
begin
 if not Visible then Exit;
 ShowWindow(Application.Handle,SW_RESTORE);
 SetFocus;
 if FoForm.Preview.Checked then
  SetWindowPos(PreviewForm.Handle,HWND_TOPMOST,0,0,0,0,SWP_NOSIZE or SWP_NOMOVE or SWP_NOACTIVATE);
end;

procedure TInGameText.ApplicationEvents1Deactivate(Sender: TObject);
begin
 if not Visible then Exit;
 if FoForm.Preview.Checked then
  begin
   SetWindowPos(PreviewForm.Handle,Handle,0,0,0,0,SWP_NOSIZE or SWP_NOMOVE or SWP_NOACTIVATE);
   SetWindowPos(Handle,PreviewForm.Handle,0,0,0,0,SWP_NOSIZE or SWP_NOMOVE or SWP_NOACTIVATE);
  end;
end;

procedure TInGameText.SaveAndCompileClick(Sender: TObject);
var s:string;
begin
 if not CheckCError then Exit;
 WriteTexts;
 if FileExists(NewName) then DeleteFile(NewName);
 if RunCompPrg('') then
  begin
   if RunCompiler.ErrorCode<>0 then
    begin
     s:='The TTDPatch in-game text compiler reported an error';
     if DatUpdated then
      begin
       s:=s+', but it managed to compile a new .dat file anyway. You will be able to use the new in-game texts.';
       ModalResult:=mrOK;
      end
     else
      s:=s+' and couldn''t create a new .dat file. Your changes were saved, but you won''t see them when running TTDPatch.';
     RunLog.InfoText.Caption:=s;
     RunLog.Memo1.Text:=RunCompiler.ProgOutput;
     RunLog.ShowModal;
    end
   else ModalResult:=mrOK;
   {if (ModalResult=mrOK) and CheckForNewFile then
    begin
     ModalResult:=mrNone;
     //ReadTexts;
    end;}
  end;
end;

function TInGameText.CheckCError: boolean;
{Check if the current cell in the string grid is correct. If it isn't, we show an error message
and return false.}
var s:WideString;
begin
 with StringGrid1 do
  if not FoForm.CompileCStyleString(Cells[Col,Row],s) then
   begin
    result:=false;
    Application.MessageBox(PChar('The C-style string you entered has an error:'#10#13+string(s)+#10#13'You must fix this error to continue or press ESC to undo your changes'),'Error in string',MB_ICONERROR);
   end
  else
   result:=true;
end;

procedure TInGameText.StringGrid1SelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
 if not CheckCError then {If there's an error in the current cell, we don't allow to leave it}
  begin
   CanSelect:=false;
   Exit;
  end;
 if CanSelect then
  with StringGrid1 do
    OrigString:=Cells[ACol,ARow];
end;

procedure TInGameText.StringGrid1KeyPress(Sender: TObject; var Key: Char);
begin
 with StringGrid1 do
  if (Key=#13) and EditorMode and not CheckCError then
  {The user wants to exit editor mode by pressing Enter, but there's an error in the text,
   so we can't allow pressing Enter}
   Key:=#0
  else if (Key=#13) and not EditorMode then
  {The user enters editor mode, but we save the original text first, so we can restore it
   when the user presses Esc}
   OrigString:=Cells[Col,Row]
  else if Key=#27 then
   begin
    {The user presses Esc - we restore the original text}
    Cells[Col,Row]:=OrigString;
    EditorMode:=false;
   end;
end;

procedure TInGameText.StringGrid1Exit(Sender: TObject);
begin
 if not CheckCError then StringGrid1.SetFocus; {If there's error in the current cell, the string grid
                                                prevents losing focus}
end;
(*
function TInGameText.CheckForNewFile:boolean;
begin
 Result:=false;
 if FileExists(NewName) then
  case Application.MessageBox('The TTDPatch in-game text compiler created an updated text file'#10#13+
                              'that contains entries that were omitted from the original txt file.'#10#13+
                              'Do you want to use this updated file instead of your older one?'#10#13+
                              '(Yes is recommended)','Updated file found',MB_ICONQUESTION or MB_YESNO) of
   ID_YES: If not ((DeleteFile(TxtName)) and (RenameFile(NewName,TxtName))) then
            Application.MessageBox('File rename failed','Error',MB_ICONERROR)
           else Result:=true;
   ID_NO: If not DeleteFile(NewName) then
           Application.MessageBox('File delete failed','Error',MB_ICONERROR);
  end;
end;  *)


procedure TInGameText.WMSysCommand(var Msg: TWMSysCommand);
begin
 if msg.CmdType=SC_MINIMIZE then
  begin
   SetWindowPos(Application.handle,Handle,Left,Top,Width,0,SWP_SHOWWINDOW);
   ShowWindow(Application.handle,SW_MINIMIZE);
  end
 else
  inherited;
end;

procedure TInGameText.HelpClick(Sender: TObject);
begin
 Application.HelpCommand(HELP_CONTEXT,10);
end;

class function TInGameText.UnQuoteString(quoted: string;
  out resultOrError: string): boolean;
var k,value,error:integer;
begin
   Result:=false;
   quoted:=trim(quoted);
   if (quoted='') or (quoted[1]<>'"') then
    begin
     resultOrError:= 'Text not in quotes'; Exit;
    end;
   Delete(quoted,1,1);
   k:=LastDelimiter('"',quoted);
   if k=0 then
    begin
     resultOrError:='Text not in quotes'; exit;
    end;
   SetLength(quoted,k-1);
   resultOrError:='';
   k:=1;
   while k<=length(quoted) do
    begin
     if quoted[k]='\' then
      begin
       if k=length(quoted) then
        begin
         resultOrError:='Backslash without control char at the end of the line'; Exit;
        end
       else
        case quoted[k+1] of
         '"': begin resultOrError:=resultOrError+'"'; inc(k,2); end;
         'n': begin resultOrError:=resultOrError+#13; inc(k,2); end;
         '\': begin resultOrError:=resultOrError+'\'; inc(k,2); end;
         else
          begin
           Val('$'+Copy(quoted,k+1,2),value,Error);
           if Error<>0 then
            begin
             resultOrError:='Incorrect backslash-sequence'; Exit;
            end;
           resultOrError:=resultOrError+chr(value);
           inc(k,3);
          end;
        end;
      end
     else
      begin
       resultOrError:=resultOrError+quoted[k]; inc(k);
      end;
    end;
 Result:=true;
end;

{ TPatchText }

constructor TPatchText.Create(ID: integer; name:string; origText: WideString);
begin
 FID:= ID;
 FName:= name;
 FOriginal:= origText;
 FTranslated:= origText;
end;

function TPatchText.isChanged: boolean;
begin
 Result:= FOriginal = FTranslated;
end;

class procedure TInGameText.ParsePatchTextList(const output: string;
  textList: TObjectList);
var currLine: string;
    lineStart,lineEnd:integer;
    id_str,name,quotedtext,text:string;
    wideText:WideString;
    i,id:integer;
    currText:TPatchText;
begin
 lineStart:=1;
 while lineStart<=length(output) do
  begin
   lineEnd:=PosEx(#10,output,lineStart);
   if lineEnd=0 then lineEnd:=length(output);
   currLine:=trim(Copy(output,lineStart,lineEnd-lineStart+1));
   lineStart:=lineEnd+1;
   if currline='' then continue;
   i:=1;
   id_str:='';
   while (i<length(currLine)) and not (currline[i] in [' ',#9]) do
    begin
     id_str:=id_str+currline[i];
     inc(i);
    end;
   while (i<length(currLine)) and (currline[i] in [' ',#9]) do
    inc(i);
   name:='';
   while (i<length(currLine)) and not (currline[i] in [' ',#9]) do
    begin
     name:=name+currline[i];
     inc(i);
    end;
   quotedText:=copy(currline,i,MaxInt);
   id:=StrToInt('$'+trim(id_str));
   UnQuoteString(quotedtext,text);
   if (text<>'') and (text[length(text)]=#0) then SetLength(text,length(text)-1);
   wideText:=FoForm.TTDTextToWideString(text);
   currText:=TPatchText.Create(id,name,wideText);
   textList.Add(currText);
  end;
end;

function TInGameText.GetPatchTexts: TObjectList;
var textList: TObjectList;
begin
 Result:=nil;
 if patchPath='' then exit;
 PatchDir:=ExtractFilePath(PatchPath);
 TxtName:=PatchDir+'ttdpttxt.txt';
 NewName:=PatchDir+'ttdpttxt.new';
 DatName:=PatchDir+'ttdpttxt.dat';
 ExeName:=PatchDir+'mkpttxt.exe';
 If not FileExists(ExeName) then
  begin
   Application.MessageBox('Your TTD directory (The directory of the TTDPatch executable) doesn''t contain mkpttxt.exe.'#10#13+
                          'This file is needed to edit TTDPatch in-game texts.'#10#13+
                          'You can download it from the TTDPatch home page (http://www.ttdpatch.net).',
                          'No in-game text compiler found',MB_ICONEXCLAMATION);
   Exit;
  end;
 {if FileExists(NewName) then
  begin
   if FileExists(TxtName) then
    begin
     s:='Your TTDPatch directory contains both ttdpttxt.txt and ttdpttxt.new'#10#13+
        'The .new file should contain new entries that was omitted from the .txt file';
     if DateTimeToFileDate(FileAge(TxtName))>DateTimeToFileDate(FileAge(NewName)) then
      s:=s+','#10#13'but the .txt file is more recent';
     s:=s+'.'#10#13'Do you want to use the .new file?'#10#13+
        '(Cancel will keep both files and close the in-game translator window)';
     case Application.MessageBox(PChar(s),'Updated file found',MB_ICONQUESTION or MB_YESNOCANCEL) of
      ID_YES: If not ((DeleteFile(TxtName)) and (RenameFile(NewName,TxtName))) then
               begin
                Application.MessageBox('File rename failed','Error',MB_ICONERROR);
                ModalResult:=mrCancel;
                Exit;
               end;
      ID_NO: If not DeleteFile(NewName) then
               begin
                Application.MessageBox('File delete failed','Error',MB_ICONERROR);
                ModalResult:=mrCancel;
                Exit;
               end;
      ID_CANCEL: begin
                  ModalResult:=mrCancel;
                  Exit;
                 end;
     end;
    end
   else RenameFile(NewName,TxtName);
  end;}
 textList := TObjectList.Create;
 if not RunCompPrg('-l') or (RunCompiler.ErrorCode<>0) then
  begin
   Application.MessageBox('An error happened while extracting known textIDs. You may continue,'+
                          'but you won''t see original texts and the GRF Export function won''t'+
                          'work correctly',
                          'Error',MB_ICONERROR or MB_OK);
  end
 else
  ParsePatchTextList(RunCompiler.ProgOutput,textList);
 //Memo1.Text:=RunCompiler.ProgOutput;
 {if FileExists(NewName) then DeleteFile(NewName);
 RunCompiler.parameters:='';
 if RunCompPrg then
  begin
   if RunCompiler.ErrorCode<>0 then
    begin
     s:='The TTDPatch in-game text compiler reported an error';
     if FileExists(TxtName) then
      s:=s+', but you may continue because none of the necessary files are missing.'
     else
      begin
       s:=s+' and couldn''t create a necessary file (ttdpttxt.txt). You cannot translate in-game texts.';
       ModalResult:=mrCancel;
      end;
     RunLog.InfoText.Caption:=s;
     RunLog.Memo1.Text:=RunCompiler.ProgOutput;
     RunLog.ShowModal;
     if ModalResult<>mrNone then Exit;
    end;
  end
 else
  begin
   if not FileExists(TxtName) then
    begin
     Application.MessageBox('Can''t find a necessary file (ttdpttxt.txt). You cannot translate in-game texts.','Error',
                            MB_ICONERROR);
     ModalResult:=MrCancel;
     Exit;
    end;
  end;}
 //CheckForNewFile;
 if (not FileExists(TxtName)) or ReadTexts(textList) then
  Result:=textList
 else
  textList.Free;
end;

procedure TInGameText.LoadPatchTexts(textList: TObjectList);
var i:integer;
begin
 StringGrid1.RowCount:=textList.Count+1;
 FormResize(Self);
 for i:=0 to textList.Count-1 do
  with textList[i] as TPatchText do
   begin
    if ID<>-1 then
     StringGrid1.Cells[0,i+1]:=IntToHex(ID,4)
    else
     StringGrid1.Cells[0,i+1]:='???';
    StringGrid1.Cells[1,i+1]:=Name;
    StringGrid1.Cells[2,i+1]:=FoForm.MakeCStyleString(Original);
    StringGrid1.Cells[3,i+1]:=FoForm.MakeCStyleString(Translated);
   end;
 PreviewTimer.Enabled:=FoForm.Preview.Checked;
 if FoForm.Preview.Checked then
  begin
   EnableWindow(PreviewForm.Handle,true);
   SetWindowPos(PreviewForm.Handle,HWND_TOPMOST,0,0,0,0,SWP_NOSIZE or SWP_NOMOVE or SWP_NOACTIVATE);
  end;
end;

procedure TInGameText.SaveChangesClick(Sender: TObject);
var txt:TPatchText;
    i,id:integer;
    s:string;
    orig, trans:WideString;
begin
 if not CheckCError then Exit;
 DestList.Clear;
 for i:=1 to StringGrid1.RowCount-1 do
  begin
   s:=StringGrid1.Cells[0,i];
   if s='???' then
    id:=-1
   else
    id:=StrToInt('$'+s);
   if not FoForm.CompileCStyleString(StringGrid1.Cells[2,i],orig) then
    orig:='';
   if not FoForm.CompileCStyleString(StringGrid1.Cells[3,i],trans) then
    trans:='';
   txt:=TPatchText.Create(id,StringGrid1.Cells[1,i],orig);
   txt.Translated:=trans;
   DestList.Add(txt);
  end;
 ModalResult:=mrOK;
end;

procedure TInGameText.LoadFromList(var msg: TMessage);
begin
 SaveAndCompile.Hide;
 SaveChanges.Show;
 GRFLabel.Show;
 LoadPatchTexts(DestList);
end;

end.

